/*
 * Copyright (C) 2021  Topview Optronics Inc.
 *
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or
 * (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License along
 * with this program; if not, write to the Free Software Foundation, Inc.,
 * 51 Franklin Street, Fifth Floor, Boston, MA 02110-1301 USA.
 */

#include <linux/module.h>
#include <linux/fs.h>
#include <linux/miscdevice.h>
#include <linux/slab.h>
#include <linux/uaccess.h>
#include <linux/device.h>
#include <linux/ctype.h>
#include <linux/delay.h>
#include <linux/gpio.h>
#include <include/lens.h>
#include <linux/string.h>
#include "lens_gpio.h"
#include "lens_motor.h"

#define MAX(a, b) ((a > b) ? (a) : (b))
#define MIN(a, b) ((a < b) ? (a) : (b))

#define BUF_LEN 64

#ifdef _MECHANISM_BOX_
static char model_type[BUF_LEN] = "box";
#else
#ifdef _MECHANISM_FIX_
static char model_type[BUF_LEN] = "fix";
#else
#ifdef _MECHANISM_DOME_
static char model_type[BUF_LEN] = "dome";
#else
#ifdef _MECHANISM_BULLET_
static char model_type[BUF_LEN] = "bullet";
#else
static char model_type[BUF_LEN] = "other";
#endif
#endif
#endif
#endif
/* Focus */
int focus_step_gpio; /* Step signal */
int focus_active_gpio; /* Active signal, unused */
int focus_max = 2640;//914; /* FIXME: Steps at far side */
int focus_min = 0; /* FIXME: Steps at near side */

/* Zoom */
int zoom_pi_gpio; /* PI signal */
int zoom_step_gpio; /* Step signal */
int zoom_active_gpio; /* Active signal, unused */
int zoom_max = 1423;//709; /* FIXME: Steps at wide side */
int zoom_min = 0; /* FIXME: Steps at tele side */

/* PIRIS */
int piris_in3_gpio;
int piris_in4_gpio;
int piris_in5_gpio;
int piris_in6_gpio;

int reset_gpio;

module_param_string(model_type, model_type , BUF_LEN, 0);
MODULE_PARM_DESC(model_type, "Model type (ex: dome, box...etc)");

/* Saved lens status */
static s32 focus_steps = 0;
module_param(focus_steps, int, 0644);
MODULE_PARM_DESC(focus_steps, "Saved current Focus Steps");

static s32 focus_pi_steps = 0;
module_param(focus_pi_steps, int, 0644);
MODULE_PARM_DESC(focus_pi_steps, "Saved Focus PI Steps");

static s32 zoom_steps = 0;
module_param(zoom_steps, int, 0644);
MODULE_PARM_DESC(zoom_steps, "Saved current Zoom Steps");

static s32 zoom_pi_steps = 0;
module_param(zoom_pi_steps, int, 0644);
MODULE_PARM_DESC(zoom_pi_steps, "Saved Zoom PI Steps");

struct lens_control {
    /* misc char device */
    struct miscdevice mdev;
    /* spi driver */
    struct pl022_device* spi;
    /* motor driver */
    struct motor_driver motor;
};

static struct lens_control* lens = NULL;
/* lens status */
/* steps increment: -1 reverse, +1 forward */
static int focus_inc = 0;
static int zoom_inc = 0;

static char is_fixed(void)
{
    if(!strncmp("fixed",model_type,5))
        return 1;
    else
        return 0;
}

static s32 lens_control_open(struct inode* inode, struct file* filp) {
    return 0;
}

static s32 lens_control_release(struct inode* inode, struct file* filp) {
    return 0;
}

static ssize_t lens_control_write(struct file* filp, const char __user* buf,
        size_t count, loff_t* ppos) {
    return count;
}

static ssize_t lens_control_read(struct file* filp, char __user* buf,
        size_t count, loff_t* ppos) {
    return count;
}

static void __always_inline wait_focus_stop(void) {

    while (read_gpio(focus_active_gpio)) {
        msleep(1);
    }
    //printk(KERN_INFO "%s() exit...\n", __func__);
}

static void __always_inline wait_zoom_stop(void) {
    while (read_gpio(zoom_active_gpio)) {
        msleep(1);
    }
    //printk(KERN_INFO "%s() exit...\n", __func__);
}

static void relative_move(struct lens_steps* steps) {
    /* CH34 Zoom */
    if (steps->zoom < 0) {
        zoom_inc = DIRECTION_ZOOM_OUT;
        lens->motor.stepping[1].reverse(abs(steps->zoom));
    } else if (steps->zoom > 0) {
        zoom_inc = DIRECTION_ZOOM_IN;
        lens->motor.stepping[1].forward(abs(steps->zoom));
    }

    /* CH12 Focus */
    if (steps->focus < 0) {
        focus_inc = DIRECTION_FOCUS_FAR;
        lens->motor.stepping[0].reverse(abs(steps->focus));
    } else if (steps->focus > 0) {
        focus_inc = DIRECTION_FOCUS_NEAR;
        lens->motor.stepping[0].forward(abs(steps->focus));
    }
}

static long lens_control_ioctl(struct file* filp, u32 cmd, void *arg) {
    int ret = 0;
    struct lens_steps steps = { 0, 0};
    int pi_bitmask = 0;
    int speed;
    int piris_step = 0;
    int type = 0;
    int value, read_value, voltage, gain;

//    printk("[PT5111]%s L%d cmd = %x======\n", __FUNCTION__, __LINE__, (unsigned int)cmd);
    switch (cmd) {
        case LENS_IOC_RELATIVE_FOCUS_MOVE:
            steps.zoom = 0;
            ret = copy_from_user(&steps.focus, arg, sizeof(steps.focus));
            printk("[PT5111]%s L%d cmd=LENS_IOC_RELATIVE_FOCUS_MOVE focus=%d\n", __FUNCTION__, __LINE__, steps.focus);
            if ((!ret) && (steps.focus != 0)) {
                relative_move(&steps);
            }
            break;
        case LENS_IOC_RELATIVE_ZOOM_MOVE:
            steps.focus = 0;
            ret = copy_from_user(&steps.zoom, arg, sizeof(steps.zoom));
            printk("[PT5111]%s L%d cmd=LENS_IOC_RELATIVE_ZOOM_MOVE zoom=%d\n", __FUNCTION__, __LINE__, steps.zoom);
            if ((!ret) && (steps.zoom != 0)) {
                relative_move(&steps);
            }
            break;
        case LENS_IOC_PIRIS_OPEN:
            ret = copy_from_user(&piris_step, arg, sizeof(piris_step));
            printk("[PT5111]%s L%d piris_step=%d\n", __FUNCTION__, __LINE__, piris_step);
            if (!ret) {
                lens->motor.piris.open(piris_step);
            }
            break;
        case LENS_IOC_PIRIS_CLOSE:
            ret = copy_from_user(&piris_step, arg, sizeof(piris_step));
            printk("[PT5111]%s L%d piris_step=%d\n", __FUNCTION__, __LINE__, piris_step);
            if (!ret) {
                lens->motor.piris.close(piris_step);
            }
            break;
        case LENS_IOC_DCIRIS_SETUP:
            printk("[PT5111]%s L%d cmd=LENS_IOC_DCIRIS_SETUP\n", __FUNCTION__, __LINE__);
            clear_gpio(piris_in3_gpio);
            clear_gpio(piris_in4_gpio);
            clear_gpio(piris_in5_gpio);
            clear_gpio(piris_in6_gpio);
            printk(KERN_DEBUG "-------------------------->>>>>>DCIRIS_SETUP<<<<<<\n");
            break;
        case LENS_IOC_FOCUS_MOVE_SPEED:
            ret = copy_from_user(&speed, arg, sizeof(speed));
            printk("[PT5111]%s L%d cmd=LENS_IOC_FOCUS_MOVE_SPEED speed=%d\n", __FUNCTION__, __LINE__, speed);

            if (!ret)
                lens->motor.stepping[0].set_speed(speed);
            break;
        case LENS_IOC_ZOOM_MOVE_SPEED:
            ret = copy_from_user(&speed, arg, sizeof(speed));
            printk("[PT5111]%s L%d cmd=LENS_IOC_ZOOM_MOVE_SPEED speed=%d\n", __FUNCTION__, __LINE__, speed);

            if (!ret)
                lens->motor.stepping[1].set_speed(speed);
            break;
        case LENS_IOC_STOP_MOVE:
            ret = copy_from_user(&type, arg, sizeof(type));
            printk("[PT5111]%s L%d cmd=LENS_IOC_STOP_MOVE type=%d\n", __FUNCTION__, __LINE__, type);

            if(!ret){
                if(type == TYPE_FOCUS)
                    lens->motor.stepping[0].stop();
                else if(type == TYPE_ZOOM)
                    lens->motor.stepping[1].stop();
            }
            break;
        case LENS_IOC_GET_FOCUS_STATUS:
            steps.focus = read_gpio(focus_active_gpio);
            printk("[PT5111]%s L%d cmd=LENS_IOC_GET_FOCUS_STATUS focus=%d\n", __FUNCTION__, __LINE__, steps.focus);
            ret = copy_to_user(arg, &steps.focus, sizeof(steps.focus));
            break;
        case LENS_IOC_GET_ZOOM_STATUS:
            steps.zoom = read_gpio(zoom_active_gpio);
            printk("[PT5111]%s L%d cmd=LENS_IOC_GET_FOCUS_STATUS zoom=%d\n", __FUNCTION__, __LINE__, steps.zoom);
            ret = copy_to_user(arg, &steps.zoom, sizeof(steps.zoom));
            break;
        case LENS_IOC_SET_ZOOM_VOLTAGE:
            ret = copy_from_user(&value, arg, sizeof(value));

            gain = (value >> 16) & 1;
            voltage = (value & 0xFFFF);


            value = (voltage << 6) | (gain << 5) | (voltage << 1);
            printk("[PT5111]%s L%d cmd=LENS_IOC_SET_ZOOM_VOLTAGE voltage=%d\n", __FUNCTION__, __LINE__, value);

            read_value = lens->motor.stepping[1].get_voltage();
            if ((!ret) && (read_value != value))
            {
                lens->motor.stepping[1].set_voltage(value);
            }
            break;
        case LENS_IOC_SET_FOCUS_VOLTAGE:
            ret = copy_from_user(&value, arg, sizeof(value));

            gain = (value >> 16) & 1;
            voltage = (value & 0xFFFF);

            value = (voltage << 6) | (gain << 4);
            printk("[PT5111]%s L%d cmd=LENS_IOC_SET_FOCUS_VOLTAGE voltage=%d\n", __FUNCTION__, __LINE__, value);

            read_value = lens->motor.stepping[0].get_voltage();
            if ((!ret) && (read_value != value))
            {
                lens->motor.stepping[0].set_voltage(value);
            }
            break;
        case LENS_IOC_SET_PIRIS_VOLTAGE:
            ret = copy_from_user(&value, arg, sizeof(value));

            voltage = ((value & 0xF) << 4);
            printk("[PT5111]%s L%d cmd=LENS_IOC_SET_PIRIS_VOLTAGE voltage=%d\n", __FUNCTION__, __LINE__, voltage);
            read_value = lens->motor.piris.get_voltage();
            if ((!ret) && (read_value != voltage))
            {
                lens->motor.piris.set_voltage(voltage);
            }
            break;

        case LENS_IOC_SET_PI:
            ret = copy_from_user(&pi_bitmask, arg, sizeof(pi_bitmask));
            printk("[PT5111]%s L%d cmd=LENS_IOC_SET_PI pi_bitmask=%d\n", __FUNCTION__, __LINE__, pi_bitmask);
            if (!ret) {
                lens->motor.pi.set(pi_bitmask);
            }
            break;

        default:
            ret = -ENOIOCTLCMD;
            printk("[PT5111]%s L%d cmd = %x ret=%d\n", __FUNCTION__, __LINE__, (unsigned int)cmd, ret);
            break;
    }
    return ret;
}

static long lens_control_unlocked_ioctl(struct file* filp, u32 cmd, unsigned long arg) {
    if (cmd == LENS_IOC_PIRIS_SETUP) {
        struct piris_setup_data pirisdata;
        int ret = copy_from_user(&pirisdata, (void *)arg, sizeof(struct piris_setup_data));
        printk("[PT5111]%s L%d ret=%d\n", __FUNCTION__, __LINE__, ret);
        set_gpio(piris_in3_gpio);
        set_gpio(piris_in4_gpio);
        set_gpio(piris_in5_gpio);
        set_gpio(piris_in6_gpio);

        lens->motor.piris.setup(pirisdata.phaseCount,(char *)pirisdata.phaseTable);
        return ret;
    }
    return lens_control_ioctl(filp, cmd, (void *)arg);
}

struct piris_setup_data1 {
    unsigned char phaseCount;
};

struct piris_setup_data2 {
    char phaseTable[8][4];
};

struct piris_setup_data3 {
    unsigned long sizeOfTable;
};

static long lens_control_compat_ioctl(struct file* filp, u32 cmd, unsigned long arg) {
    if (cmd == LENS_IOC_PIRIS_SETUP) {
        struct piris_setup_data pirisdata;
        int ret = copy_from_user(&pirisdata, (void *)arg, sizeof(struct piris_setup_data));
        printk("[PT5111]%s L%d ret=%d\n", __FUNCTION__, __LINE__, ret);
        set_gpio(piris_in3_gpio);
        set_gpio(piris_in4_gpio);
        set_gpio(piris_in5_gpio);
        set_gpio(piris_in6_gpio);

        lens->motor.piris.setup(pirisdata.phaseCount,(char *)pirisdata.phaseTable);
        return ret;
    }
    return lens_control_ioctl(filp, cmd, (void *)arg);
}

static const struct file_operations lens_control_fops = {
    .open    = lens_control_open,
    .release = lens_control_release,
    .write   = lens_control_write,
    .read    = lens_control_read,
    .unlocked_ioctl   = lens_control_unlocked_ioctl,
    .compat_ioctl = lens_control_compat_ioctl,
};

irqreturn_t focus_step_cb(int irq, void *data) {
    focus_steps = MAX(MIN((focus_steps + focus_inc), focus_max), focus_min);

    return IRQ_HANDLED;
}

irqreturn_t zoom_pi_cb(int irq, void *data) {
    //zoom_pi_level = read_gpio(zoom_pi_gpio);
    //printk("[PT5111]zoom pi callback %d\n", zoom_pi_level);

    return IRQ_HANDLED;
}

irqreturn_t zoom_step_cb(int irq, void *data) {
    zoom_steps = MAX(MIN((zoom_steps + zoom_inc), zoom_max), zoom_min);

    return IRQ_HANDLED;
}

s32 __init lens_control_init(void) {
    s32 ret = 0;

    printk(KERN_INFO "PT5111 model_type: %s\n", model_type);
    if(is_fixed()) {
        lens = kzalloc(sizeof(struct lens_control), GFP_KERNEL);
        if (!lens) {
            ret = -ENOMEM;
            goto End;
        }
        /* initialize misc driver */
        lens->mdev.minor = MISC_DYNAMIC_MINOR;
        lens->mdev.name = "lens";
        lens->mdev.fops = &lens_control_fops;
        ret = misc_register(&lens->mdev);

//        setup_gpio_direction(ir_cut, GPIO_OUTPUT);
//        clear_gpio(ir_cut);
    } else {

        lens = kzalloc(sizeof(struct lens_control), GFP_KERNEL);
        if (!lens) {
            ret = -ENOMEM;
            goto End;
        }

        /* initialize motor active signal */
#define SETUP_ACTIVE_SIGNAL_GPIO(name) \
        setup_gpio_direction(name, GPIO_INPUT)
        SETUP_ACTIVE_SIGNAL_GPIO(focus_active_gpio);
        SETUP_ACTIVE_SIGNAL_GPIO(zoom_active_gpio);
        SETUP_ACTIVE_SIGNAL_GPIO(focus_step_gpio);
        //SETUP_ACTIVE_SIGNAL_GPIO(zoom_pi_gpio);
        SETUP_ACTIVE_SIGNAL_GPIO(zoom_step_gpio);
#undef SETUP_ACTIVE_SIGNAL_GPIO

//        setup_gpio_direction(ir_cut, GPIO_OUTPUT);
        setup_gpio_direction(piris_in3_gpio, GPIO_OUTPUT);
        setup_gpio_direction(piris_in4_gpio, GPIO_OUTPUT);
        setup_gpio_direction(piris_in5_gpio, GPIO_OUTPUT);
        setup_gpio_direction(piris_in6_gpio, GPIO_OUTPUT);

        if (gpio_is_valid(reset_gpio))
        {
            setup_gpio_direction(reset_gpio, GPIO_OUTPUT);
            set_gpio(reset_gpio);
            mdelay(2);
            clear_gpio(reset_gpio);
            mdelay(2);
        }

        ret = open_motor_driver(NULL, &lens->motor, MOTOR_DRIVER_PT5111);
        if (ret != 0) {
            goto End;
        }

        set_gpio(piris_in3_gpio);
        set_gpio(piris_in4_gpio);
        set_gpio(piris_in5_gpio);
        set_gpio(piris_in6_gpio);

        //zoom_pi_level = read_gpio(zoom_pi_gpio);
        //printk(KERN_INFO "current zoom pi %d\n", zoom_pi_level);

        /* initialize misc driver */
        lens->mdev.minor = MISC_DYNAMIC_MINOR;
        lens->mdev.name = "lens";
        lens->mdev.fops = &lens_control_fops;
        ret = misc_register(&lens->mdev);
        if (ret < 0) {
            goto End;
        }

End:
        if (ret != 0) {
            /* failure, release allocated resources */
            if (lens != NULL) {
                kfree(lens);
            }
        } else {

            /* initialize interrupt */
#define SETUP_GPIO_INTERRUPT(name, mode) \
            setup_gpio_interrupt(name##_gpio, mode, name##_cb)
            //SETUP_GPIO_INTERRUPT(focus_pi, TRIGGER_DUAL_EDGE);
            SETUP_GPIO_INTERRUPT(focus_step, TRIGGER_UP_EDGE);
            //SETUP_GPIO_INTERRUPT(zoom_pi, TRIGGER_DUAL_EDGE);
            SETUP_GPIO_INTERRUPT(zoom_step, TRIGGER_UP_EDGE);
#undef SETUP_GPIO_INTERRUPT

            /* success, enable all interrupts */
#define ENABLE_GPIO_INTERRUPT(name) \
            enable_gpio_interrupt(name##_gpio)
            //ENABLE_GPIO_INTERRUPT(focus_pi);
            ENABLE_GPIO_INTERRUPT(focus_step);
            //ENABLE_GPIO_INTERRUPT(zoom_pi);
            ENABLE_GPIO_INTERRUPT(zoom_step);
#undef ENABLE_GPIO_INTERRUPT
        }
    }
    return ret;
}

void __exit lens_control_exit(void) {
    if (lens != NULL) {
        if(is_fixed()) {
//            free_gpio(ir_cut);
        } else {
            /* disable all interrupts */
#define DISABLE_GPIO_INTERRUPT(name) \
            disable_gpio_interrupt(name##_gpio)
            //DISABLE_GPIO_INTERRUPT(focus_pi);
            DISABLE_GPIO_INTERRUPT(focus_step);
            //DISABLE_GPIO_INTERRUPT(zoom_pi);
            DISABLE_GPIO_INTERRUPT(zoom_step);
#undef DISABLE_GPIO_INTERRUPT

//            free_gpio(ir_cut);

            free_gpio(piris_in3_gpio);
            free_gpio(piris_in4_gpio);
            free_gpio(piris_in5_gpio);
            free_gpio(piris_in6_gpio);
        }
        misc_deregister(&lens->mdev);
        kfree(lens);
    }
}
