/*
 * Copyright (C) 2021  Topview Optronics Inc.
 *
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or
 * (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License along
 * with this program; if not, write to the Free Software Foundation, Inc.,
 * 51 Franklin Street, Fifth Floor, Boston, MA 02110-1301 USA.
 */

#include <linux/module.h>
#include <linux/delay.h>
#include <linux/mutex.h>
#include <linux/device.h>
#include <linux/of.h>
#include <linux/of_gpio.h>
#include <linux/regmap.h>
#include <linux/spi/spi.h>

#include "lens_motor.h"
#include "pt5111_define.h"
#include "lens_gpio.h"

extern s32 __init lens_control_init(void);
extern void __exit lens_control_exit(void);

extern int focus_step_gpio;
extern int focus_active_gpio;
extern int zoom_step_gpio;
extern int zoom_active_gpio;
extern int piris_in3_gpio;
extern int piris_in4_gpio;
extern int piris_in5_gpio;
extern int piris_in6_gpio;
extern int reset_gpio;

static struct spi_device *pt5111_spi_dev;

/*
static char piris_spi[8][4]={{0, 1, 1, 0}, // refer to PT51111 SPI spec, page #13, truth table of IN3, 4, 5 and 6
                               {1, 1, 1, 0},
                               {1, 0, 1, 0},
                               {1, 0, 1, 1},
                               {1, 0, 0, 1},
                               {1, 1, 0, 1},
                               {0, 1, 0, 1},
                               {0, 1, 1, 1}};
*/

static char piris_custom_table[8][4] =
{
    {0, 0, 0, 0},
    {0, 0, 0, 0},
    {0, 0, 0, 0},
    {0, 0, 0, 0},
    {0, 0, 0, 0},
    {0, 0, 0, 0},
    {0, 0, 0, 0},
    {0, 0, 0, 0}
};
static int piris_phase_count = 8;
static int piris_is_setuped = 0;
static int cur_phase = 0;

static u16 zoom_power = 0x0;
static u16 focus_power = 0x0;

static struct mutex motor_lock;

struct pt5111_device {
    u32 speed[2];
    u32 voltage[3];
};

static struct pt5111_device pt5111_dev = {
    .speed = { 469, 469 },
    //0x310 0x338 5.2V Gain two times
    //0x200 0x210 3.4V Gain one time
    .voltage = {0x200, 0x210, VOLT_230mv},
};

/*****************************************************************************/
/*                             private function                              */
/*****************************************************************************/
static void __always_inline set_speed(u32 idx, u32 speed) 
{
    pt5111_dev.speed[idx] = speed;
}

static void __always_inline set_voltage(u32 idx, u32 voltage) {
    pt5111_dev.voltage[idx] = voltage;
}

static u32 __always_inline get_voltage(u32 idx)
{
    return pt5111_dev.voltage[idx];
}

static int __always_inline motor_spi_write(unsigned short cmd) 
{
    u8 tx[2] = {0};

    tx[0] = (cmd & 0xff00) >> 8;
    tx[1] = cmd & 0xff;

    return spi_write(pt5111_spi_dev, tx, sizeof(tx));
}

static void __always_inline stop(void) 
{
    motor_spi_write(COMMON | PS_CTRL | PS_ON | zoom_power | focus_power);
}

/*****************************************************************************/
/*                          motor driver function                            */
/*****************************************************************************/
static void stepping0_forward(u32 steps);

static s32 initialize(void *ptr) 
{

    mutex_init(&motor_lock);

    motor_spi_write(COMMON | SW_RST | ALL_RESET);  /* software reset */
    motor_spi_write(COMMON | EXT_OUT | EXT_EXCIT | EXT_MOB1_EXT1_EXT2);

    /* setup channel 1/2 for stepping moter */
    motor_spi_write(CHANNEL12 | DIRECTION | MICRO_MODE | NO_ACC_NO_DEC);
    motor_spi_write(CHANNEL12 | PULSE_RATE_RANGE | PRR_02u | MICRO256);
    motor_spi_write(CHANNEL12 | PULSE_RATE0 | pt5111_dev.speed[0]);
    motor_spi_write(CHANNEL12 | PRE_EXCIT | EXCITIME_0);
    motor_spi_write(CHANNEL12 | POST_EXCIT | EXCITIME_0);
    //spi_write(CHANNEL12 | POST_EXCIT | EXCITIME_6);
    //spi_write(CHANNEL12 | VOLTAGE);// FUll Scale Voltage
    motor_spi_write(CHANNEL12 | VOLTAGE | pt5111_dev.voltage[0]);

    /* setup channel 3/4 for stepping moter */
    motor_spi_write(CHANNEL34 | DIRECTION | MICRO_MODE | NO_ACC_NO_DEC);
    motor_spi_write(CHANNEL34 | PULSE_RATE_RANGE | PRR_02u | MICRO256);
    motor_spi_write(CHANNEL34 | PULSE_RATE0 | pt5111_dev.speed[1]);
    motor_spi_write(CHANNEL34 | PRE_EXCIT | EXCITIME_0);
    motor_spi_write(CHANNEL34 | POST_EXCIT | EXCITIME_0);
    //spi_write(CHANNEL34 | POST_EXCIT | EXCITIME_6);
    //spi_write(CHANNEL34 | VOLTAGE); // FUll Scale Voltage
    motor_spi_write(CHANNEL34 | VOLTAGE | pt5111_dev.voltage[1]);

    motor_spi_write(CHANNEL57 |DIRECTION|CH5_CC); //PIRIS_DRIVING_MODE
    motor_spi_write(COMMON |INPUT_CTRL|IN_CTRL|CH5_TWO_WIRE|CH5_IN_IN|CH6_TWO_WIRE|CH6_IN_IN); //PIRIS_INPUT_PING_CTRL

    /* setup defualt piris voltage (channel5/6) */
    motor_spi_write(COMMON | CHAN5_CTRL | pt5111_dev.voltage[2] );
    motor_spi_write(COMMON | CHAN6_CTRL | pt5111_dev.voltage[2] );

    stepping0_forward(0); /* FIXME: initial pt5111 */
    //motor_spi_write(COMMON | PS_CTRL | PS_OFF);
    motor_spi_write(COMMON | PS_CTRL | PS_ON);

    printk(KERN_INFO "[PT5111 INIT]\n");
    return 0;
}

static void release(void) 
{
    memset(&pt5111_dev, 0x00, sizeof(pt5111_dev));
}

static void stepping0_speed(u32 speed) 
{
    set_speed(0, speed);
}

static void stepping0_voltage(u32 voltage) {
    set_voltage(0, voltage);
    motor_spi_write(CHANNEL12 | VOLTAGE | pt5111_dev.voltage[0]);
}

static u32 stepping0_get_voltage(void)
{
  return get_voltage(0);
}

#define STEPPING_MOVE(channel, direction, speed, steps) \
    motor_spi_write(COMMON | PS_CTRL | PS_ON | zoom_power | focus_power); \
motor_spi_write(CHANNEL##channel | DIRECTION | direction | MICRO_MODE | NO_ACC_NO_DEC); \
motor_spi_write(CHANNEL##channel | PULSE_RATE0 | speed); \
motor_spi_write(CHANNEL##channel | NO_PULSES1 | (steps & 0x03FF)); \
motor_spi_write(CHANNEL##channel | NO_PULSES2 | ((steps >> 10) & 0x03FF)); \
motor_spi_write(CHANNEL##channel | OPERATION_CTRL | ACTIVE | EXCIT_ON)


static void stepping0_forward(u32 steps) 
{
    mutex_lock(&motor_lock);
    focus_power = PS12_ON;
    STEPPING_MOVE(12, FORWARD, pt5111_dev.speed[0], steps);
    mutex_unlock(&motor_lock);
}

static void stepping0_reverse(u32 steps) 
{
    mutex_lock(&motor_lock);
    focus_power = PS12_ON;
    STEPPING_MOVE(12, REVERSE, pt5111_dev.speed[0], steps);
    mutex_unlock(&motor_lock);
}

static void stepping0_stop(void) 
{
    mutex_lock(&motor_lock);
    focus_power = 0x0;
    stop();
    mutex_unlock(&motor_lock);
}

static void stepping1_speed(u32 speed) 
{
    set_speed(1, speed);
}

static void stepping1_voltage(u32 voltage) {
    set_voltage(1, voltage);
    motor_spi_write(CHANNEL34 | VOLTAGE | pt5111_dev.voltage[1]);
}

static u32 stepping1_get_voltage(void)
{
  return get_voltage(1);
}

static void stepping1_forward(u32 steps) 
{
    mutex_lock(&motor_lock);
    zoom_power = PS34_ON;
    STEPPING_MOVE(34, FORWARD, pt5111_dev.speed[1], steps);
    mutex_unlock(&motor_lock);
}

static void stepping1_reverse(u32 steps) 
{
    mutex_lock(&motor_lock);
    zoom_power = PS34_ON;
    STEPPING_MOVE(34, REVERSE, pt5111_dev.speed[1], steps);
    mutex_unlock(&motor_lock);
}

static void stepping1_stop(void) 
{
    mutex_lock(&motor_lock);
    zoom_power = 0x0;
    stop();
    mutex_unlock(&motor_lock);
}

static void piris_voltage(u32 voltage) {
    set_voltage(2, voltage);
    motor_spi_write(COMMON | CHAN5_CTRL | pt5111_dev.voltage[2] );
    motor_spi_write(COMMON | CHAN6_CTRL | pt5111_dev.voltage[2] );
}

static u32 piris_get_voltage(void)
{
    return get_voltage(2);
}

static void set_daynight(enum daynight daynight) 
{
#define DN_SWITCH(direction) \
    /* set channel 5 for ICR (daynight) */ \
    motor_spi_write(CHANNEL57 | DIRECTION | CH5_FS); \
    motor_spi_write(CHANNEL57 | PWM57_SETTING | direction | PWM5_STOP_BRAKE); \
    motor_spi_write(CHANNEL57 | PWM5_DUTY | 0x0300); \
    motor_spi_write(CHANNEL57 | OPERATION_CTRL | ACTIVE | EXCIT_ON); \
    msleep(10); \
    motor_spi_write(CHANNEL57 | OPERATION_CTRL | STOP | EXCIT_OFF)

    switch (daynight) {
        case DAY_MODE:
            printk(KERN_INFO "[lens] set to day mode\n");
            DN_SWITCH(PWM5_REVERSE);
            break;
        case NIGHT_MODE:
            printk(KERN_INFO "[lens] set to night mode\n");
            DN_SWITCH(PWM5_FORWARD);
            break;
    }

#undef DN_SWITCH
}

static void set_pi(int bitmask) 
{
    int i;
    unsigned short pi = 0;

    for (i = 0; i < 3; i++) {
        int on = (bitmask >> i) & 1;
        pi |= ((on == 1) ? (PI1_ON) : (PI1_OFF)) >> i;
        printk(KERN_INFO "[PT5111] set PI%d %s\n", (i + 1), (on ? "on" : "off"));
    }

    motor_spi_write(COMMON | PI_CTRL | pi);

}

static void control_gpio(int gpio_id, int direction)
{
    if(direction)
        set_gpio(gpio_id);
    else
        clear_gpio(gpio_id);
}

//static void pt5111_ps_on(void)
//{
//    motor_spi_write(COMMON | PS_CTRL | PS_ON);
//}
//
//static void pt5111_ps_off(void)
//{
//    motor_spi_write(COMMON | PS_CTRL | PS_OFF);
//}


//set 1-2 3-4 channel power on/off
/*
   static u32 pt5111_power_status = 0;
   static void pt5111_ps_on(){
   pt5111_power_status &= PS_ON;
   motor_spi_write(COMMON | PS_CTRL | pt5111_power_status);
   }

   static void pt5111_ps_off(){
   pt5111_power_status &= PS_OFF;
   motor_spi_write(COMMON | PS_CTRL | pt5111_power_status);
   }

   static void pt5111_ps12_on(){

   }
   static void pt5111_ps12_off() {

   }
   static void pt5111_ps34_on(){

   }
   static void pt5111_ps34_off(){

   }
 */

//static void piris_phase_off(void)
//{
//    control_gpio(piris_in3_gpio, 0);
//    control_gpio(piris_in4_gpio, 0);
//    control_gpio(piris_in5_gpio, 0);
//    control_gpio(piris_in6_gpio, 0);
//}


static void piris_phase_break(void)
{
    control_gpio(piris_in3_gpio, 1);
    control_gpio(piris_in4_gpio, 1);
    control_gpio(piris_in5_gpio, 1);
    control_gpio(piris_in6_gpio, 1);
}


static void operate_piris(int cur_phase)
{
    if(piris_is_setuped) {

        control_gpio(piris_in3_gpio, piris_custom_table[cur_phase][0]);
        control_gpio(piris_in4_gpio, piris_custom_table[cur_phase][1]);
        control_gpio(piris_in5_gpio, piris_custom_table[cur_phase][2]);
        control_gpio(piris_in6_gpio, piris_custom_table[cur_phase][3]);


    } else {

        control_gpio(piris_in3_gpio, piris_custom_table[cur_phase][0]);
        control_gpio(piris_in4_gpio, piris_custom_table[cur_phase][1]);
        control_gpio(piris_in5_gpio, piris_custom_table[cur_phase][2]);
    }

    //delay time for control motor pps or use printk to set delay time
    msleep(5);
    /*
       printk("%s: phase=%d\n","piris", cur_phase+1);
       printk("%s: IN3=%d\n","piris", piris_custom_table[cur_phase][0]);
       printk("%s: IN4=%d\n","piris", piris_custom_table[cur_phase][1]);
       printk("%s: IN5=%d\n","piris", piris_custom_table[cur_phase][2]);
       printk("%s: IN6(inverted)=%d\n","piris", piris_custom_table[cur_phase][3]);
    //printk("%s: Out6A(A+)=%d\n","piris", piris_phase[cur_phase][0]);
    //printk("%s: Out5A(B+)=%d\n","piris", piris_phase[cur_phase][1]);
    //printk("%s: Out6B(A-)=%d\n","piris", piris_phase[cur_phase][2]);
    //printk("%s: Out5B(B-)=%d\n\n","piris", piris_phase[cur_phase][3]);
     */
}

static void piris_open(u32 step)
{
    int i;

    printk("%s: step=%d\n","piris_open", step);

    for (i=0; i < step; i++) {// open p-iris
        operate_piris(cur_phase);
        cur_phase++;
        if(cur_phase > piris_phase_count -1 ) cur_phase = 0;
    }
    piris_phase_break();
}

static void piris_close(u32 step)
{
    int i;

    printk("%s: step=%d\n","piris_close", step);

    for (i=0; i < step; i++) {// close p-iris
        operate_piris(cur_phase);
        cur_phase--;
        if(cur_phase < 0) cur_phase = piris_phase_count-1;
    }
    piris_phase_break();
}


static void piris_setup(int phasecount, char *phasetable)
{
    int i,j;
    cur_phase = 0;
    piris_phase_count = phasecount;
    piris_is_setuped = 1;

    memcpy(piris_custom_table, phasetable, sizeof(piris_custom_table));

    printk("[piris_setup] piris_phase_count = %d\n",piris_phase_count);
    printk("[piris_setup] piris_phase_table:\n");
    for(i=0; i < piris_phase_count; i++) {
        for(j=0; j<4; j++) {
            printk("%d ",piris_custom_table[i][j]);
        }
        printk("\n");
    }
    printk("[piris_setup] EXIT piris_setup\n");

}

struct motor_driver pt5111 = {
    .initialize = initialize,
    .release = release,
    .piris = {
        .open = piris_open,
        .close = piris_close,
        .setup = piris_setup,
        .set_voltage = piris_voltage,
        .get_voltage = piris_get_voltage,
    },
    .stepping = {
        {
            .set_speed = stepping0_speed,
            .set_voltage = stepping0_voltage,
            .get_voltage = stepping0_get_voltage,
            .forward = stepping0_forward,
            .reverse = stepping0_reverse,
            .stop = stepping0_stop,
        },
        {
            .set_speed = stepping1_speed,
            .set_voltage = stepping1_voltage,
            .get_voltage = stepping1_get_voltage,
            .forward = stepping1_forward,
            .reverse = stepping1_reverse,
            .stop = stepping1_stop,
        },

    },
    .daynight = {
        .set = set_daynight,
    },
    .pi = {
        .set = set_pi,
    },
};

static int pt5111_probe(struct spi_device *spi)
{
    pt5111_spi_dev = spi;

    focus_step_gpio = of_get_named_gpio(spi->dev.of_node, "focus-step", 0);
    focus_active_gpio = of_get_named_gpio(spi->dev.of_node, "focus-active", 0);
    zoom_step_gpio = of_get_named_gpio(spi->dev.of_node, "zoom-step", 0);
    zoom_active_gpio = of_get_named_gpio(spi->dev.of_node, "zoom-active", 0);
    piris_in3_gpio = of_get_named_gpio(spi->dev.of_node, "piris-in3", 0);
    piris_in4_gpio = of_get_named_gpio(spi->dev.of_node, "piris-in4", 0);
    piris_in5_gpio = of_get_named_gpio(spi->dev.of_node, "piris-in5", 0);
    piris_in6_gpio = of_get_named_gpio(spi->dev.of_node, "piris-in6", 0);
    reset_gpio = of_get_named_gpio(spi->dev.of_node, "pt5111-reset", 0);

    if (!gpio_is_valid(focus_step_gpio)) {
        dev_err(&spi->dev, "invalid focus-step gpio (%d)\n", focus_step_gpio);
        return -ENODEV;
    }
    if (!gpio_is_valid(focus_active_gpio)) {
        dev_err(&spi->dev, "invalid focus-active gpio (%d)\n", focus_active_gpio);
        return -ENODEV;
    }
    if (!gpio_is_valid(zoom_step_gpio)) {
        dev_err(&spi->dev, "invalid zoom-step gpio (%d)\n", zoom_step_gpio);
        return -ENODEV;
    }
    if (!gpio_is_valid(zoom_active_gpio)) {
        dev_err(&spi->dev, "invalid zoom-active gpio (%d)\n", zoom_active_gpio);
        return -ENODEV;
    }
    if (!gpio_is_valid(piris_in3_gpio)) {
        dev_err(&spi->dev, "invalid piris-in3 gpio (%d)\n", piris_in3_gpio);
        return -ENODEV;
    }
    if (!gpio_is_valid(piris_in4_gpio)) {
        dev_err(&spi->dev, "invalid piris-in4 gpio (%d)\n", piris_in4_gpio);
        return -ENODEV;
    }
    if (!gpio_is_valid(piris_in5_gpio)) {
        dev_err(&spi->dev, "invalid piris-in5 gpio (%d)\n", piris_in5_gpio);
        return -ENODEV;
    }
    if (!gpio_is_valid(piris_in6_gpio)) {
        dev_err(&spi->dev, "invalid piris-in6 gpio (%d)\n", piris_in6_gpio);
        return -ENODEV;
    }
    if (!gpio_is_valid(reset_gpio) && reset_gpio != -ENOENT) {
        dev_err(&spi->dev, "invalid pt5111-reset gpio (%d)\n", reset_gpio);
        return -ENODEV;
    }

    lens_control_init();

    return 0;
}

static int pt5111_remove(struct spi_device *spi)
{
    lens_control_exit();

    return 0;
}

static const struct spi_device_id pt5111_spi_id[] = {
    {"pt5111", 0},
    {}
};
MODULE_DEVICE_TABLE(spi, pt5111_spi_id);

static const struct of_device_id pt5111_spi_match[] = {
    { .compatible = "pt5111"},
    { },
};
MODULE_DEVICE_TABLE(of, pt5111_spi_match);

static struct spi_driver pt5111_spi_driver = {
    .driver = {
        .name = "pt5111",
        .of_match_table = pt5111_spi_match,
    },
    .probe = pt5111_probe,
    .remove = pt5111_remove,
    .id_table = pt5111_spi_id,
};

module_spi_driver(pt5111_spi_driver);

MODULE_AUTHOR("Topview Optronics Corp.");
MODULE_DESCRIPTION("Lens Control");
MODULE_LICENSE("GPL v2");

